@@ -28,24 +28,32 @@ from base64 import standard_b64encode
 
 import bz2
 import os
+import shutil
+import tempfile
 import tarfile
 import zipfile
 
 from ....revision import (
+    NULL_REVISION,
     Revision,
     )
 from ....tests import (
     TestCase,
     )
-from ....tests.features import PluginLoadedFeature
+from ....errors import NoSuchFile
+from ....tests.features import (
+    ModuleAvailableFeature,
+    PluginLoadedFeature,
+    )
 from ..config import (
     DebBuildConfig,
     )
 from ..errors import (
     MissingUpstreamTarball,
     PackageVersionNotPresent,
-    WatchFileMissing,
     )
+
+from ....sixish import text_type
 from . import (
     LzmaFeature,
     XzFeature,
@@ -59,6 +67,7 @@ from ..upstream import (
     UpstreamProvider,
     UpstreamSource,
     UScanSource,
+    UScanError,
     extract_tarball_version,
     gather_orig_files,
     new_tarball_name,
@@ -67,6 +76,7 @@ from ..util import (
     component_from_orig_tarball,
     )
 from ..upstream.branch import (
+    guess_upstream_revspec,
     get_export_upstream_revision,
     get_snapshot_revision,
     LazyUpstreamBranchSource,
@@ -78,11 +88,14 @@ from ..upstream.branch import (
 from ..upstream.pristinetar import (
     PristineTarSource,
     is_upstream_tag,
+    revision_pristine_tar_format,
+    revision_pristine_tar_delta,
     upstream_tag_version,
     )
 
 
 svn_plugin = PluginLoadedFeature('svn')
+dulwich = ModuleAvailableFeature('dulwich')
 
 
 class MockSources(object):
@@ -148,9 +161,10 @@ class MockAptCaller(object):
 class AptSourceTests(TestCase):
 
     def test_get_apt_command_for_source(self):
-        self.assertEqual("apt-get source -y --only-source --tar-only "
-                "apackage=someversion",
-                AptSource()._get_command("apackage", "someversion"))
+        self.assertEqual(
+            "apt-get source -y --only-source --tar-only "
+            "apackage=someversion",
+            AptSource()._get_command("apackage", "someversion"))
 
     def test_apt_provider_no_package(self):
         caller = MockAptCaller()
@@ -158,7 +172,8 @@ class AptSourceTests(TestCase):
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
-        self.assertRaises(PackageVersionNotPresent, src.fetch_tarballs,
+        self.assertRaises(
+            PackageVersionNotPresent, src.fetch_tarballs,
             "apackage", "0.2", "target", _apt_pkg=apt_pkg)
         self.assertEqual(1, apt_pkg.init_called_times)
         self.assertEqual(1, apt_pkg.get_pkg_source_records_called_times)
@@ -169,12 +184,14 @@ class AptSourceTests(TestCase):
 
     def test_apt_provider_wrong_version(self):
         caller = MockAptCaller()
-        sources = MockSources(["0.1-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")]])
+        sources = MockSources(
+            ["0.1-1"],
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
-        self.assertRaises(PackageVersionNotPresent, src.fetch_tarballs,
+        self.assertRaises(
+            PackageVersionNotPresent, src.fetch_tarballs,
             "apackage", "0.2", "target", _apt_pkg=apt_pkg)
         self.assertEqual(1, apt_pkg.init_called_times)
         self.assertEqual(1, apt_pkg.get_pkg_source_records_called_times)
@@ -186,9 +203,9 @@ class AptSourceTests(TestCase):
     def test_apt_provider_multiple_tarballs(self):
         caller = MockAptCaller(work=True)
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2.orig.tar.bz2", "tar"),
-              ("checksum", 1L, "apackage_0.2.orig-extra.tar.gz", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2.orig.tar.bz2", "tar"),
+              ("checksum", 1, "apackage_0.2.orig-extra.tar.gz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -201,8 +218,8 @@ class AptSourceTests(TestCase):
     def test_apt_provider_right_version_bz2(self):
         caller = MockAptCaller(work=True)
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2.orig.tar.bz2", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2.orig.tar.bz2", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -213,8 +230,8 @@ class AptSourceTests(TestCase):
     def test_apt_provider_right_version_xz(self):
         caller = MockAptCaller(work=True)
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2.orig.tar.xz", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2.orig.tar.xz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -225,8 +242,8 @@ class AptSourceTests(TestCase):
     def test_apt_provider_right_version(self):
         caller = MockAptCaller(work=True)
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2.orig.tar.gz", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2.orig.tar.gz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -247,8 +264,8 @@ class AptSourceTests(TestCase):
     def test_apt_provider_right_version_command_fails(self):
         caller = MockAptCaller()
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2.orig.tar.gz", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2.orig.tar.gz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -269,8 +286,8 @@ class AptSourceTests(TestCase):
     def test_apt_provider_right_version_is_native(self):
         caller = MockAptCaller(work=True)
         sources = MockSources(["0.1-1", "0.2-1"],
-            [[("checksum", 0L, "apackage_0.1.orig.tar.gz", "tar")],
-             [("checksum", 0L, "apackage_0.2-1.orig.tar.gz", "tar")]])
+            [[("checksum", 0, "apackage_0.1.orig.tar.gz", "tar")],
+             [("checksum", 0, "apackage_0.2-1.orig.tar.gz", "tar")]])
         apt_pkg = MockAptPkg(sources)
         src = AptSource()
         src._run_apt_source = caller.call
@@ -286,14 +303,25 @@ class AptSourceTests(TestCase):
 
 class RecordingSource(UpstreamSource):
 
-    def __init__(self, succeed, latest=None):
+    def __init__(self, succeed, latest=None, recent=None):
         self._succeed = succeed
         self._specific_versions = []
+        if recent:
+            recent.sort()
+        if latest is None and recent:
+            latest = recent[-1]
         self._latest = latest
+        if recent is None and latest:
+            recent = [latest]
+        self._recent = recent
 
     def get_latest_version(self, package, current_version):
         return self._latest
 
+    def get_recent_versions(self, package, since_version=None):
+        return [v for v in self._recent
+                if since_version is None or v > since_version]
+
     def fetch_tarballs(self, package, version, target_dir, components=None):
         self._specific_versions.append((package, version, target_dir))
         if not self._succeed:
@@ -322,11 +350,24 @@ class StackedUpstreamSourceTests(TestCase):
         stack = StackedUpstreamSource([a, b])
         self.assertEquals("1.1", stack.get_latest_version("mypkg", "1.0"))
 
+    def test_get_recent_versions(self):
+        a = RecordingSource(False, recent=["1.0", "1.1"])
+        b = RecordingSource(False, recent=["1.1", "1.2"])
+        stack = StackedUpstreamSource([a, b])
+        self.assertEquals(
+            ["1.0", "1.1", "1.2"], stack.get_recent_versions("mypkg", "0.9"))
+        self.assertEquals(
+            ["1.0", "1.1", "1.2"], stack.get_recent_versions("mypkg"))
+        self.assertEquals(
+            ["1.2"], stack.get_recent_versions("mypkg", "1.1"))
+
     def test_repr(self):
-        self.assertEquals("StackedUpstreamSource([])",
-                repr(StackedUpstreamSource([])))
-        self.assertEquals("StackedUpstreamSource([RecordingSource()])",
-                repr(StackedUpstreamSource([RecordingSource(False)])))
+        self.assertEquals(
+            "StackedUpstreamSource([])",
+            repr(StackedUpstreamSource([])))
+        self.assertEquals(
+            "StackedUpstreamSource([RecordingSource()])",
+            repr(StackedUpstreamSource([RecordingSource(False)])))
 
     def test_none(self):
         a = RecordingSource(False)
@@ -345,26 +386,31 @@ class UScanSourceTests(TestCaseWithTransport):
         self.tree = self.make_branch_and_tree('.')
 
     def test_export_watchfile_none(self):
-        src = UScanSource(self.tree, False)
-        self.assertRaises(WatchFileMissing, src._export_watchfile)
+        src = UScanSource(self.tree, '', False)
+        self.assertRaises(NoSuchFile, src._export_file, 'watch', self.test_dir)
 
     def test_export_watchfile_top_level(self):
-        src = UScanSource(self.tree, True)
+        src = UScanSource(self.tree, '', True)
         self.build_tree(['watch'])
-        self.assertRaises(WatchFileMissing, src._export_watchfile)
         self.tree.add(['watch'])
-        self.assertTrue(src._export_watchfile() is not None)
+        tmpdir = tempfile.mkdtemp()
+        try:
+            self.assertIsNot(src._export_file('watch', tmpdir), None)
+        finally:
+            shutil.rmtree(tmpdir)
 
     def test_export_watchfile(self):
-        src = UScanSource(self.tree, False)
+        src = UScanSource(self.tree, '', False)
         self.build_tree(['debian/', 'debian/watch'])
-        self.assertRaises(WatchFileMissing, src._export_watchfile)
         self.tree.smart_add(['debian/watch'])
-        self.assertTrue(src._export_watchfile() is not None)
+        tmpdir = tempfile.mkdtemp()
+        try:
+            self.assertIsNot(src._export_file('watch', tmpdir), None)
+        finally:
+            shutil.rmtree(tmpdir)
 
     def test__xml_report_extract_upstream_version(self):
-        self.assertEquals("1.2.9",
-            UScanSource._xml_report_extract_upstream_version("""
+        text = b"""
 <dehs>
 <package>tdb</package>
 <debian-uversion>1.2.8</debian-uversion>
@@ -372,18 +418,85 @@ class UScanSourceTests(TestCaseWithTransport):
 <upstream-version>1.2.9</upstream-version>
 <upstream-url>ftp://ftp.samba.org/pub/tdb/tdb-1.2.9.tar.gz</upstream-url>
 <status>Newer version available</status>
-</dehs>"""))
+</dehs>"""
+        self.assertEquals("1.2.9",
+            UScanSource._xml_report_extract_upstream_version(text))
 
     def test__xml_report_extract_upstream_version_warnings(self):
-        self.assertIs(None,
-            UScanSource._xml_report_extract_upstream_version("""
+        text = b"""
 <dehs>
 <package>tdb</package>
 <warnings>uscan warning: Unable to determine current version
 in debian/watch, skipping:
 ftp://ftp.samba.org/pub/tdb/tdb-(.+).tar.gz</warnings>
 </dehs>
-"""))
+"""
+        self.assertIs(
+            None,
+            UScanSource._xml_report_extract_upstream_version(text))
+
+    def test__xml_report_extract_upstream_version_noise(self):
+        text = b"""
+<dehs>
+blahf =>
+<package>tdb</package>
+<debian-uversion>1.2.8</debian-uversion>
+<upstream-version>1.2.9</upstream-version>
+<warnings>uscan warning: Unable to determine current version
+in debian/watch, skipping:
+ftp://ftp.samba.org/pub/tdb/tdb-(.+).tar.gz</warnings>
+</dehs>
+"""
+        self.assertEquals("1.2.9",
+            UScanSource._xml_report_extract_upstream_version(text))
+
+    def test__xml_report_extract_upstream_version_errors(self):
+        text = b"""
+<dehs>
+blahf =>
+<package>tdb</package>
+<debian-uversion>1.2.8</debian-uversion>
+<errors>something something signature</errors>
+</dehs>
+"""
+        self.assertRaises(
+            UScanError,
+            UScanSource._xml_report_extract_upstream_version, text)
+
+
+class GuessUpstreamRevspecTests(TestCase):
+
+    def test_guess_upstream_revspec(self):
+        self.assertEqual(
+                [u'tag:1.2', u'tag:foo-1.2', u'tag:v1.2', u'tag:v.1.2'],
+                list(guess_upstream_revspec('foo', '1.2')))
+
+    def test_snapshot(self):
+        self.assertEqual(
+                [u'revno:1200', u'tag:1.2+bzr1200', u'tag:foo-1.2+bzr1200',
+                 u'tag:v1.2+bzr1200', u'tag:v.1.2+bzr1200'],
+                list(guess_upstream_revspec('foo', '1.2+bzr1200')))
+        self.assertEqual(
+                [u'revno:1200', u'tag:1.2~bzr1200', u'tag:foo-1.2~bzr1200',
+                 u'tag:v1.2~bzr1200', u'tag:v.1.2~bzr1200'],
+                list(guess_upstream_revspec('foo', '1.2~bzr1200')))
+        self.assertEqual(
+                [u'git:abc1def', u'date:2018-01-01', u'tag:1.2~git20180101.abc1def',
+                 u'tag:foo-1.2~git20180101.abc1def',
+                 u'tag:v1.2~git20180101.abc1def',
+                 u'tag:v.1.2~git20180101.abc1def'],
+                list(guess_upstream_revspec('foo', '1.2~git20180101.abc1def')))
+        self.assertEqual(
+                [u'git:abc1def', u'date:2018-01-01', u'tag:1.2~git20180101.0.abc1def',
+                 u'tag:foo-1.2~git20180101.0.abc1def',
+                 u'tag:v1.2~git20180101.0.abc1def',
+                 u'tag:v.1.2~git20180101.0.abc1def'],
+                list(guess_upstream_revspec('foo', '1.2~git20180101.0.abc1def')))
+        self.assertEqual(
+                [u'date:2018-01-01', u'tag:1.2~git20180101',
+                 u'tag:foo-1.2~git20180101',
+                 u'tag:v1.2~git20180101', u'tag:v.1.2~git20180101'],
+                list(guess_upstream_revspec('foo', '1.2~git20180101')))
 
 
 class UpstreamBranchSourceTests(TestCaseWithTransport):
@@ -391,13 +504,13 @@ class UpstreamBranchSourceTests(TestCaseWithTransport):
 
     def setUp(self):
         super(UpstreamBranchSourceTests, self).setUp()
-        self.tree = self.make_branch_and_tree('.')
+        self.tree = self.make_branch_and_tree('upstream')
 
     def test_fetch_tarballs(self):
         self.tree.commit("msg")
         self.tree.branch.tags.set_tag("1.0", self.tree.branch.last_revision())
         source = UpstreamBranchSource(self.tree.branch,
-            {"1.0": self.tree.branch.last_revision()})
+            {"1.0": self.tree.branch.last_revision().decode('utf-8')})
         os.mkdir("mydir")
         self.assertEquals(["mydir/foo_1.0.orig.tar.gz"],
             source.fetch_tarballs("foo", "1.0", "mydir"))
@@ -413,11 +526,24 @@ class UpstreamBranchSourceTests(TestCaseWithTransport):
         self.tree.commit("msg")
         self.tree.branch.tags.set_tag("2.1", self.tree.branch.last_revision())
         source = UpstreamBranchSource(self.tree.branch,
-            {"2.1": self.tree.branch.last_revision()})
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')})
         self.assertEquals("2.1", source.get_latest_version("foo", "1.0"))
         self.tree.commit("msg")
         self.assertEquals("2.1+bzr2", source.get_latest_version("foo", "1.0"))
 
+    def test_get_recent_versions(self):
+        self.tree.commit("msg")
+        self.tree.branch.tags.set_tag("2.1", self.tree.branch.last_revision())
+        source = UpstreamBranchSource(
+            self.tree.branch,
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')})
+        self.assertEquals(["2.1"], source.get_recent_versions("foo", "1.0"))
+        revid2 = self.tree.commit("msg")
+        self.tree.branch.tags.set_tag('2.2', revid2)
+        self.assertEquals(
+            ["2.1", "2.2"],
+            source.get_recent_versions("foo", "1.0"))
+
     def test_version_as_revisions(self):
         revid1 = self.tree.commit("msg")
         self.tree.branch.tags.set_tag("2.1", self.tree.branch.last_revision())
@@ -425,12 +551,12 @@ class UpstreamBranchSourceTests(TestCaseWithTransport):
             [('user.conf', True), ('default.conf', False)],
             branch=self.tree.branch)
         source = UpstreamBranchSource(self.tree.branch,
-            {"2.1": self.tree.branch.last_revision()},
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')},
             config=config)
         revid2 = self.tree.commit("msg")
         self.assertEquals(revid2,
-            source.version_as_revision("foo", "2.1+bzr2"))
-        self.assertEquals({None: revid1}, source.version_as_revisions("foo", "2.1"))
+            source.version_as_revision("foo", u"2.1+bzr2"))
+        self.assertEquals({None: revid1}, source.version_as_revisions("foo", u"2.1"))
 
     def test_version_as_revision(self):
         revid1 = self.tree.commit("msg")
@@ -439,7 +565,7 @@ class UpstreamBranchSourceTests(TestCaseWithTransport):
             [('user.conf', True), ('default.conf', False)],
             branch=self.tree.branch)
         source = UpstreamBranchSource(self.tree.branch,
-            {"2.1": self.tree.branch.last_revision()},
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')},
             config=config)
         revid2 = self.tree.commit("msg")
         self.assertEquals(revid2,
@@ -469,10 +595,28 @@ class UpstreamBranchSourceTests(TestCaseWithTransport):
         config = DebBuildConfig(
             [('user.conf', True), ('default.conf', False)],
             branch=self.tree.branch)
-        source = UpstreamBranchSource(self.tree.branch, {"2.1": "tag:foo"},
-            config=config)
-        self.assertRaises(PackageVersionNotPresent,
-            source.version_as_revision, "foo", "2.1")
+        source = UpstreamBranchSource(
+            self.tree.branch, {"2.1": "tag:foo"}, config=config)
+        self.assertRaises(
+            PackageVersionNotPresent,
+            source.version_as_revision, "foo", u"2.1")
+
+    def test_get_latest_version_no_random_access(self):
+        local_repo = self.make_repository('local')
+        self.tree.commit("msg")
+        self.tree.branch.tags.set_tag("2.1", self.tree.branch.last_revision())
+        self.tree.branch.repository.supports_random_access = False
+        source = UpstreamBranchSource.from_branch(
+            self.tree.branch,
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')},
+            local_dir=local_repo.controldir)
+        self.assertEquals("2.1", source.get_latest_version("foo", "1.0"))
+        self.tree.commit("msg")
+        source = UpstreamBranchSource.from_branch(
+            self.tree.branch,
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')},
+            local_dir=local_repo.controldir)
+        self.assertEquals("2.1+bzr2", source.get_latest_version("foo", "1.0"))
 
 
 class LazyUpstreamBranchSourceTests(TestCaseWithTransport):
@@ -486,7 +630,7 @@ class LazyUpstreamBranchSourceTests(TestCaseWithTransport):
         self.tree.commit("msg")
         self.tree.branch.tags.set_tag("1.0", self.tree.branch.last_revision())
         source = LazyUpstreamBranchSource(self.tree.branch.base,
-            {"1.0": self.tree.branch.last_revision()})
+            {"1.0": self.tree.branch.last_revision().decode('utf-8')})
         self.assertIs(None, source._upstream_branch)
         os.mkdir("mydir")
         self.assertEquals(["mydir/foo_1.0.orig.tar.gz"],
@@ -513,6 +657,21 @@ class LazyUpstreamBranchSourceTests(TestCaseWithTransport):
         self.assertEquals("2.1+bzr2", source.get_latest_version("foo", "1.0"))
         self.assertIsNot(None, source._upstream_branch)
 
+    def test_get_recent_versions(self):
+        revid1 = self.tree.commit("msg")
+        self.tree.branch.tags.set_tag("2.1", revid1)
+        source = LazyUpstreamBranchSource(
+            self.tree.branch.base,
+            {"2.1": self.tree.branch.last_revision()})
+        self.assertIs(None, source._upstream_branch)
+        self.assertEquals(["2.1"], source.get_recent_versions("foo", "1.0"))
+        revid2 = self.tree.commit("msg")
+        self.tree.branch.tags.set_tag("2.2", revid2)
+        self.assertEquals(
+            ["2.1", "2.2"],
+            source.get_recent_versions("foo", "1.0"))
+        self.assertIsNot(None, source._upstream_branch)
+
     def test_version_as_revision(self):
         revid1 = self.tree.commit("msg")
         self.tree.branch.tags.set_tag("2.1", self.tree.branch.last_revision())
@@ -520,7 +679,7 @@ class LazyUpstreamBranchSourceTests(TestCaseWithTransport):
             [('user.conf', True), ('default.conf', False)],
             branch=self.tree.branch)
         source = LazyUpstreamBranchSource(self.tree.branch.base,
-            {"2.1": self.tree.branch.last_revision()},
+            {"2.1": self.tree.branch.last_revision().decode('utf-8')},
             config=config)
         self.assertIs(None, source._upstream_branch)
         revid2 = self.tree.commit("msg")
@@ -541,38 +700,38 @@ class TestUpstreamBranchVersion(TestCase):
         return "%s+bzr%d" % (version_string, revno)
 
     def test_snapshot_none_existing(self):
-      self.revhistory = ["somerevid"]
+      self.revhistory = [b"somerevid"]
       self.assertEquals("1.2+bzr1",
-          _upstream_branch_version(self.revhistory, {}, "bla", "1.2", self.get_suffix))
+          _upstream_branch_version(self.revhistory, b"somerevid", {}, "bla", "1.2", self.get_suffix))
 
     def test_snapshot_nothing_new(self):
       self.revhistory = []
       self.assertEquals("1.2",
-          _upstream_branch_version(self.revhistory, {}, "bla", "1.2", self.get_suffix))
+          _upstream_branch_version(self.revhistory, NULL_REVISION, {}, "bla", "1.2", self.get_suffix))
 
     def test_new_tagged_release(self):
       """Last revision is tagged - use as upstream version."""
-      self.revhistory = ["somerevid"]
+      self.revhistory = [b"somerevid"]
       self.assertEquals("1.3",
-          _upstream_branch_version(self.revhistory, {"somerevid": [u"1.3"]}, "bla", "1.2", self.get_suffix))
+          _upstream_branch_version(self.revhistory, b"somerevid", {b"somerevid": [u"1.3"]}, "bla", "1.2", self.get_suffix))
 
     def test_refresh_snapshot_pre(self):
-      self.revhistory = ["somerevid", "oldrevid"]
+      self.revhistory = [b"somerevid", b"oldrevid"]
       self.assertEquals("1.3~bzr2",
-          _upstream_branch_version(self.revhistory, {}, "bla", "1.3~bzr1",
+          _upstream_branch_version(self.revhistory, b"somerevid", {}, "bla", "1.3~bzr1",
               self.get_suffix))
 
     def test_refresh_snapshot_post(self):
-      self.revhistory = ["somerevid", "oldrevid"]
+      self.revhistory = [b"somerevid", b"oldrevid"]
       self.assertEquals("1.3+bzr2",
-          _upstream_branch_version(self.revhistory, {}, "bla", "1.3+bzr1",
+          _upstream_branch_version(self.revhistory, b"somerevid", {}, "bla", "1.3+bzr1",
               self.get_suffix))
 
     def test_new_tag_refresh_snapshot(self):
-      self.revhistory = ["newrevid", "somerevid", "oldrevid"]
+      self.revhistory = [b"newrevid", b"somerevid", b"oldrevid"]
       self.assertEquals("1.3+bzr3",
-            _upstream_branch_version(self.revhistory,
-                {"somerevid": [u"1.3"]}, "bla", "1.2+bzr1", self.get_suffix))
+            _upstream_branch_version(self.revhistory, b"newrevid",
+                {b"somerevid": [u"1.3"]}, "bla", "1.2+bzr1", self.get_suffix))
 
 
 class TestUpstreamTagToVersion(TestCase):
@@ -593,7 +752,9 @@ class TestUpstreamTagToVersion(TestCase):
         self.assertEquals("42.0", upstream_tag_to_version(u"bla-42.0", "bla"))
 
     def test_unicode(self):
-        self.assertEquals("42.0\xc2\xa9", upstream_tag_to_version("bla-42.0\xc2\xa9".decode("utf-8"), "bla"))
+        self.assertEquals(
+                u"42.0\xa9",
+                upstream_tag_to_version(u"bla-42.0\xa9", "bla"))
 
 
 class TestUpstreamVersionAddRevision(TestCaseWithTransport):
@@ -602,13 +763,17 @@ class TestUpstreamVersionAddRevision(TestCaseWithTransport):
     def setUp(self):
         super(TestUpstreamVersionAddRevision, self).setUp()
         self.revnos = {}
-        self.svn_revnos = {"somesvnrev": 45}
-        self.revnos = {"somerev": 42, "somesvnrev": 12}
+        self.svn_revnos = {b"somesvnrev": 45}
+        self.git_shas = {b"somegitrev": b"e7f47cf254a6ddd4996fe41fa6115bd32eff5437"}
+        self.revnos = {b"somerev": 42, b"somesvnrev": 12, b"somegitrev": 66}
         self.repository = self
 
     def revision_id_to_revno(self, revid):
         return self.revnos[revid]
 
+    def revision_id_to_dotted_revno(self, revid):
+        return (self.revnos[revid], )
+
     def get_revision(self, revid):
         rev = Revision(revid)
         if revid in self.svn_revnos:
@@ -617,31 +782,63 @@ class TestUpstreamVersionAddRevision(TestCaseWithTransport):
             rev.foreign_revid = ("uuid", "bp", self.svn_revnos[revid])
             from ...svn import mapping
             rev.mapping = mapping.mapping_registry.get_default()()
+        if revid in self.git_shas:
+            self.requireFeature(dulwich)
+            # Fake a bzr-svn revision
+            rev.foreign_revid = self.git_shas[revid]
+            from ....git import mapping
+            rev.mapping = mapping.mapping_registry.get_default()()
+            rev.timestamp = 1514772000
+            rev.timezone = 0
         return rev
 
     def test_update_plus_rev(self):
         self.assertEquals("1.3+bzr42",
-          upstream_version_add_revision(self, "1.3+bzr23", "somerev"))
+          upstream_version_add_revision(self, "1.3+bzr23", b"somerev"))
 
     def test_update_tilde_rev(self):
         self.assertEquals("1.3~bzr42",
-          upstream_version_add_revision(self, "1.3~bzr23", "somerev"))
+          upstream_version_add_revision(self, "1.3~bzr23", b"somerev"))
 
     def test_new_rev(self):
         self.assertEquals("1.3+bzr42",
-          upstream_version_add_revision(self, "1.3", "somerev"))
+          upstream_version_add_revision(self, "1.3", b"somerev"))
 
     def test_svn_new_rev(self):
         self.assertEquals("1.3+svn45",
-          upstream_version_add_revision(self, "1.3", "somesvnrev"))
+          upstream_version_add_revision(self, "1.3", b"somesvnrev"))
 
     def test_svn_plus_rev(self):
         self.assertEquals("1.3+svn45",
-          upstream_version_add_revision(self, "1.3+svn3", "somesvnrev"))
+          upstream_version_add_revision(self, "1.3+svn3", b"somesvnrev"))
 
     def test_svn_tilde_rev(self):
         self.assertEquals("1.3~svn45",
-            upstream_version_add_revision(self, "1.3~svn800", "somesvnrev"))
+            upstream_version_add_revision(self, "1.3~svn800", b"somesvnrev"))
+
+    def test_git_tilde_rev(self):
+        self.assertEquals(
+            "1.3~git20180101.e7f47cf",
+            upstream_version_add_revision(
+                self, "1.3~git20171201.11b1d57", b"somegitrev"))
+
+    def test_git_new_rev(self):
+        self.assertEquals(
+            "1.3+git20180101.e7f47cf",
+            upstream_version_add_revision(
+                self, "1.3", b"somegitrev"))
+        self.assertEquals(
+            "1.0~git20180101",
+            upstream_version_add_revision(
+                self, "1.0~git20160320", b"somegitrev"))
+        self.assertEquals(
+            "1.0~git20180101.1.e7f47cf",
+            upstream_version_add_revision(
+                self, "1.0~git20180101.0.11b1d57", b"somegitrev"))
+        self.assertEquals(
+            "1.0~git20180101.0.e7f47cf",
+            upstream_version_add_revision(
+                self, "1.0~git20170101.0.11b1d57", b"somegitrev"))
 
 
 class GetExportUpstreamRevisionTests(TestCase):
@@ -690,6 +887,14 @@ class GetRevisionSnapshotTests(TestCase):
     def test_with_svn_snapshot_plus(self):
         self.assertEquals("svn:2424", get_snapshot_revision("0.4.4+svn2424"))
 
+    def test_git(self):
+        self.assertEquals(
+            "date:20190101",
+            get_snapshot_revision("0.4.4+git20190101"))
+        self.assertEquals(
+            "git:abc1def",
+            get_snapshot_revision("0.4.4+git20190101.abc1def"))
+
 
 class TestIsUpstreamTag(TestCase):
 
@@ -735,63 +940,110 @@ class TestUpstreamTagVersion(TestCase):
             upstream_tag_version('upstream-debian-2.1/lib'))
 
 
+class GenericPristineTarSourceTests(TestCase):
+
+    def test_pristine_tar_format_gz(self):
+        rev = Revision("myrevid")
+        rev.properties[u"deb-pristine-delta"] = "1"
+        self.assertEquals("gz", revision_pristine_tar_format(rev))
+
+    def test_pristine_tar_format_bz2(self):
+        rev = Revision("myrevid")
+        rev.properties[u"deb-pristine-delta-bz2"] = "1"
+        self.assertEquals("bz2", revision_pristine_tar_format(rev))
+
+    def test_pristine_tar_format_xz(self):
+        rev = Revision("myrevid")
+        rev.properties[u"deb-pristine-delta-xz"] = "1"
+        self.assertEquals("xz", revision_pristine_tar_format(rev))
+
+    def test_pristine_tar_format_unknown(self):
+        rev = Revision("myrevid")
+        self.assertRaises(AssertionError,
+            revision_pristine_tar_format, rev)
+
+    def test_pristine_tar_delta_unknown(self):
+        rev = Revision("myrevid")
+        self.assertRaises(AssertionError,
+            revision_pristine_tar_delta, rev)
+
+    def test_pristine_tar_delta_gz(self):
+        rev = Revision("myrevid")
+        rev.properties[u"deb-pristine-delta"] = standard_b64encode(b"bla")
+        self.assertEquals(b"bla", revision_pristine_tar_delta(rev))
+
+
 class PristineTarSourceTests(TestCaseWithTransport):
 
     def setUp(self):
         super(PristineTarSourceTests, self).setUp()
         self.tree = self.make_branch_and_tree('unstable')
-        root_id = self.tree.path2id("")
-        self.source = PristineTarSource(self.tree, self.tree.branch)
+        self.source = PristineTarSource(self.tree.branch)
+
+    def test_gbp_tag_format(self):
+        self.build_tree_contents([
+            ('unstable/debian/', ),
+            ('unstable/debian/gbp.conf', """\
+[DEFAULT]
+upstream-tag = blah-%(version)s
+""")])
+        self.source = PristineTarSource.from_tree(
+            self.tree.branch, self.tree)
+        upstream_v_no = "0.1"
+        self.assertEqual(
+            self.source.tag_name(upstream_v_no),
+            "blah-" + upstream_v_no)
+
+    def test_gbp_tag_format_unusual(self):
+        self.build_tree_contents([
+            ('unstable/debian/', ),
+            ('unstable/debian/gbp.conf', """\
+[DEFAULT]
+upstream-tag = blah-%(version%~%-)s
+""")])
+        self.source = PristineTarSource.from_tree(
+            self.tree.branch, self.tree)
+        self.assertEqual(
+            self.source.tag_name("0.1~1"),
+            "blah-0.1-1")
 
     def test_upstream_tag_name(self):
         upstream_v_no = "0.1"
-        self.assertEqual(self.source.tag_name(upstream_v_no),
-                "upstream-" + upstream_v_no)
+        self.assertEqual(
+            self.source.tag_name(upstream_v_no),
+            "upstream-" + upstream_v_no)
 
     def test_tag_name_distro(self):
-        self.assertEquals(self.source.tag_name("0.3", distro="ubuntu"),
-                "upstream-ubuntu-0.3")
+        self.assertEquals(
+            self.source.tag_name("0.3", distro="ubuntu"),
+            "upstream-ubuntu-0.3")
 
     def test_version(self):
-        self.assertEquals(['upstream-3.3', 'upstream-debian-3.3',
-            'upstream-ubuntu-3.3', 'upstream/3.3', 'upstream_3.3'],
+        self.assertEquals([
+            'upstream-3.3',
+            'upstream-debian-3.3',
+            'upstream-ubuntu-3.3', 'upstream/3.3', 'upstream-3.3',
+            'upstream_3.3'],
             self.source.possible_tag_names("3.3", component=None))
 
+    def test_version_with_tilde(self):
+        self.assertEquals([
+            'upstream-3.3~brz232',
+            'upstream-debian-3.3~brz232',
+            'upstream-ubuntu-3.3~brz232',
+            'upstream/3.3~brz232',
+            'upstream-3.3~brz232',
+            'upstream/3.3_brz232',
+            'upstream_3.3~brz232'],
+            self.source.possible_tag_names("3.3~brz232", component=None))
+
     def test_version_component(self):
-        self.assertEquals(['upstream-3.3/extlib', 'upstream-debian-3.3/extlib',
+        self.assertEquals([
+            'upstream-3.3/extlib',
+            'upstream-debian-3.3/extlib',
             'upstream-ubuntu-3.3/extlib'],
             self.source.possible_tag_names("3.3", component="extlib"))
 
-    def test_pristine_tar_format_gz(self):
-        rev = Revision("myrevid")
-        rev.properties["deb-pristine-delta"] = "1"
-        self.assertEquals("gz", self.source.pristine_tar_format(rev))
-
-    def test_pristine_tar_format_bz2(self):
-        rev = Revision("myrevid")
-        rev.properties["deb-pristine-delta-bz2"] = "1"
-        self.assertEquals("bz2", self.source.pristine_tar_format(rev))
-
-    def test_pristine_tar_format_xz(self):
-        rev = Revision("myrevid")
-        rev.properties["deb-pristine-delta-xz"] = "1"
-        self.assertEquals("xz", self.source.pristine_tar_format(rev))
-
-    def test_pristine_tar_format_unknown(self):
-        rev = Revision("myrevid")
-        self.assertRaises(AssertionError,
-            self.source.pristine_tar_format, rev)
-
-    def test_pristine_tar_delta_unknown(self):
-        rev = Revision("myrevid")
-        self.assertRaises(AssertionError,
-            self.source.pristine_tar_delta, rev)
-
-    def test_pristine_tar_delta_gz(self):
-        rev = Revision("myrevid")
-        rev.properties["deb-pristine-delta"] = standard_b64encode("bla")
-        self.assertEquals("bla", self.source.pristine_tar_delta(rev))
-
     def test_version_as_revisions_missing(self):
         self.assertRaises(PackageVersionNotPresent,
             self.source.version_as_revisions, None, "1.2")
@@ -866,14 +1118,18 @@ class TarfileSourceTests(TestCaseWithTransport):
 
     def test_version_unicode_not_specified(self):
         source = TarfileSource(u"foo-1.0.tar.gz")
-        latest_version = source.get_latest_version("foo", "0.9")
+        latest_version = source.get_latest_version("foo", u"0.9")
         self.assertEquals("1.0", latest_version)
-        self.assertIsInstance(latest_version, str)
+        self.assertIsInstance(latest_version, text_type)
 
     def test_get_latest_version_parses(self):
         source = TarfileSource("foo-1.0.tar.gz")
         self.assertEquals("1.0", source.get_latest_version("foo", "0.9"))
 
+    def test_get_recent_versions(self):
+        source = TarfileSource("foo-1.0.tar.gz")
+        self.assertEquals(["1.0"], source.get_recent_versions("foo", "0.9"))
+
     def test_fetch_tarballs(self):
         source = TarfileSource("foo-1.0.tar.gz", "1.0")
         os.mkdir("bar")
@@ -950,9 +1206,8 @@ class _TouchUpstreamProvider(UpstreamProvider):
 
     def provide(self, target_dir):
         path = os.path.join(target_dir, self.desired_tarball_name)
-        f = open(path, "wb")
-        f.write("I am a tarball, honest\n")
-        f.close()
+        with open(path, "wb") as f:
+            f.write(b"I am a tarball, honest\n")
         return [path]
 
 
@@ -1026,11 +1281,8 @@ class TestGatherOrigTarballs(TestCaseWithTransport):
             set(gather_orig_files("mypkg", "1.0", ".")))
 
     def test_utf8_invalid_file(self):
-        f = open("\xf6o.tar.gz", "w")
-        try:
+        with open("\xf6o.tar.gz", "w") as f:
             f.write("foo")
-        finally:
-            f.close()
         self.build_tree(["mypkg_1.0.orig.tar.gz"])
         self.assertEquals(
             [os.path.join(self.test_dir, "mypkg_1.0.orig.tar.gz")],
